<?php

/**
 * Zend Framework (http://framework.zend.com/)
 *
 * @link      http://github.com/zendframework/Reservation for the canonical source repository
 * @copyright Copyright (c) 2005-2015 Zend Technologies USA Inc. (http://www.zend.com)
 * @license   http://framework.zend.com/license/new-bsd New BSD License
 */

namespace Reservation\Controller;

use Zend\Mvc\Controller\AbstractRestfulController;
use Zend\Http\Response;
use Zend\View\Model\JsonModel;
use Zend\Http\Request;
use Log\Model\LogEntity;

class ReservationController extends AbstractRestfulController {

    public function deleteList($data) {
        $request = $this->getRequest();

        // Check if authorised
        $isAuthorised = $this->AuthorisationController()->isAuthorisationValid($request);
        if(!$isAuthorised) {
            return $this->emptyResponse($this->getResponse(), 403);
        }

        // Create mappers
        /** @var $reservationMapper \Reservation\Model\ReservationMapper */
        $reservationMapper = $this->getMapper('ReservationMapper');
        /** @var $quantityMapper \Quantity\Model\QuantityMapper */
        $quantityMapper = $this->getMapper('QuantityMapper');

        // Get quantity from data
        $reservationToken = (isset($data['reservation_token']) ? $data['reservation_token'] : null);

        // Error if reservation token isn't given
        if (!$reservationToken) {
            return $this->emptyResponse($this->getResponse(), 400);
        }

        // Get token expiry
        $reservationTokenExpiry = $reservationMapper->getReservationTokenExpiry($reservationToken);

        // If expiry is null, it means the token doesn't exist so we return a 404
        if ($reservationTokenExpiry === null) {
            return $this->emptyResponse($this->getResponse(), 404);
        }

        $this->addLog($request, '', $reservationToken, 'Reservation Cleared');

        // Remove all items for previous token
        $reservationMapper->clearReservationsByToken($reservationToken, $quantityMapper);

        return new JsonModel(array(

        ));
    }

    public function update($productNumber, $data) {
        $request = $this->getRequest();

        // Check if authorised
        $isAuthorised = $this->AuthorisationController()->isAuthorisationValid($request);
        if(!$isAuthorised) {
            return $this->emptyResponse($this->getResponse(), 403);
        }

        // Create mappers
        /** @var $reservationMapper \Reservation\Model\ReservationMapper */
        $reservationMapper = $this->getMapper('ReservationMapper');
        /** @var $quantityMapper \Quantity\Model\QuantityMapper */
        $quantityMapper = $this->getMapper('QuantityMapper');

        // Get quantity from data
        $requestedQuantity = (isset($data['requested_quantity']) ? $data['requested_quantity'] : -1);
        $reservationToken = (isset($data['reservation_token']) ? $data['reservation_token'] : null);

        // Error if quantity is < 0
        if ($requestedQuantity < 0) {
            return $this->emptyResponse($this->getResponse(), 400);
        }

        // Check reservation token is given and not expired
        $newTokenNeeded = false;
        if ($reservationToken == null) {
            $newTokenNeeded = true;
        } else {
            // Get token expiry
            $reservationTokenExpiry = $reservationMapper->getReservationTokenExpiry($reservationToken);

            // If expiry is null, it means the token doesn't exist so we return a 404
            if ($reservationTokenExpiry === null) {
                return $this->emptyResponse($this->getResponse(), 404);
            }

            // Check if it's already expired
            if ($reservationTokenExpiry <= date('Y-m-d H:i:s')) {
                // Expired so we will set the response code to 408
                $response = $this->getResponse();
                $response->setStatusCode(408);

                // Remove all items for previous token
                $reservationMapper->clearReservationsByToken($reservationToken, $quantityMapper);

                // And flag to create a new reservation token
                $newTokenNeeded = true;

                $this->addLog($request, '', $reservationToken, 'Token expired, clearing reservation');
            }
        }

        // Generate a new token if needed
        if ($newTokenNeeded) {
            $reservationTokenDetails = $reservationMapper->generateToken();
            $reservationToken = $reservationTokenDetails['token'];
            $reservationTokenExpiry = $reservationTokenDetails['token_expiry'];
        }

        // Get the current quantity for the product
        $quantity = $quantityMapper->getQuantityByProductNumber($productNumber);

        // Check quantity is valid (null if product doesn't exist)
        if($quantity === null) {
            return $this->emptyResponse($this->getResponse(), 404);
        }

        // Get current quantity reserved for product and token
        $currentReservedQuantity = $reservationMapper->getReservedQuantityByProductNumberAndReservationToken($productNumber, $reservationToken);

        // Check whether there is sufficient quantity to fulfil the request
        $isSufficientQuantity = false;
        if ($requestedQuantity <= $currentReservedQuantity) {
            // Check if reducing quantity (requesting less than currently reserved)
            $isSufficientQuantity = true;
        } else {
            // If we're increasing quantity, we need to take off the amount that has currently been reserved
            if ($requestedQuantity <= ($quantity['available_quantity'] + $currentReservedQuantity)) {
                $isSufficientQuantity = true;
            }
        }

        // Check available quantity is greater than or equal to the requested quantity
        if (!$isSufficientQuantity) {
            $this->addLog($request, $productNumber, $reservationToken, 'Not enough quantity, requested: ' . $requestedQuantity . ', available: ' . $quantity['available_quantity']);
            $this->getResponse()->setStatusCode(409);
            return new JsonModel(array(
                'total_quantity' => $quantity['total_quantity'],
                'available_quantity' => $quantity['available_quantity'],
            ));
        }

        // Update the reserved quantity for the product
        $isReserved = $reservationMapper->setReservationQuantity($productNumber, $requestedQuantity, $reservationToken, $quantityMapper);

        // If it wasn't reserved, something went wrong somewhere so just throw an internal server error
        if (!$isReserved) {
            return $this->emptyResponse($this->getResponse(), 500);
        }

        // Get all reservations for current token
        $reservations = $reservationMapper->getAllReservationsByToken($reservationToken);

        $this->addLog($request, $productNumber, $reservationToken, 'Reserved ' . $requestedQuantity . ' successfully. Total reserved is ' . $requestedQuantity);

        return new JsonModel(array(
            'products' => $reservations,
            'reservation_token' => $reservationToken,
            'expiry_datetime' => $reservationTokenExpiry
        ));
    }

    public function patch($productNumber, $data) {
        $request = $this->getRequest();

        // Check if authorised
        $isAuthorised = $this->AuthorisationController()->isAuthorisationValid($request);
        if(!$isAuthorised) {
            return $this->emptyResponse($this->getResponse(), 403);
        }

        // Create mappers
        /** @var $reservationMapper \Reservation\Model\ReservationMapper */
        $reservationMapper = $this->getMapper('ReservationMapper');
        /** @var $quantityMapper \Quantity\Model\QuantityMapper */
        $quantityMapper = $this->getMapper('QuantityMapper');

        // Get quantity from data
        $requestedQuantity = (isset($data['requested_quantity']) ? $data['requested_quantity'] : -1);
        $reservationToken = (isset($data['reservation_token']) ? $data['reservation_token'] : null);

        // Error if quantity is < 1
        if ($requestedQuantity < 1) {
            return $this->emptyResponse($this->getResponse(), 400);
        }

        // Error if reservation token isn't given
        if (!$reservationToken) {
            return $this->emptyResponse($this->getResponse(), 400);
        }

        // Get token expiry
        $reservationTokenExpiry = $reservationMapper->getReservationTokenExpiry($reservationToken);

        // If expiry is null, it means the token doesn't exist so we return a 404
        if ($reservationTokenExpiry === null) {
            return $this->emptyResponse($this->getResponse(), 404);
        }

        // Check if it's already expired
        $newTokenNeeded = false;
        if ($reservationTokenExpiry <= date('Y-m-d H:i:s')) {
            // Expired so we will set the response code to 408
            $response = $this->getResponse();
            $response->setStatusCode(408);

            // Remove all items for previous token
            $reservationMapper->clearReservationsByToken($reservationToken, $quantityMapper);

            // And flag to create a new reservation token
            $newTokenNeeded = true;

            $this->addLog($request, '', $reservationToken, 'Token expired, clearing reservation');
        }

        // Generate a new token if needed
        if ($newTokenNeeded) {
            $reservationTokenDetails = $reservationMapper->generateToken();
            $reservationToken = $reservationTokenDetails['token'];
            $reservationTokenExpiry = $reservationTokenDetails['token_expiry'];
        }

        // Get the current quantity for the product
        $quantity = $quantityMapper->getQuantityByProductNumber($productNumber);

        // Check quantity is valid (null if product doesn't exist)
        if($quantity === null) {
            return $this->emptyResponse($this->getResponse(), 404);
        }

        // Get current quantity reserved for product and token
        $currentReservedQuantity = $reservationMapper->getReservedQuantityByProductNumberAndReservationToken($productNumber, $reservationToken);

        // Check available quantity is greater than or equal to the requested quantity
        if ($requestedQuantity > $quantity['available_quantity']) {
            $this->addLog($request, $productNumber, $reservationToken, 'Not enough quantity, requested: ' . $requestedQuantity . ', available: ' . $quantity['available_quantity']);
            $this->getResponse()->setStatusCode(409);
            return new JsonModel(array(
                'total_quantity' => $quantity['total_quantity'],
                'available_quantity' => $quantity['available_quantity'],
            ));
        }

        // Update the reserved quantity for the product
        $isReserved = $reservationMapper->increaseReservationQuantity($productNumber, $requestedQuantity, $reservationToken, $quantityMapper);

        // If it wasn't reserved, something went wrong somewhere so just throw an internal server error
        if (!$isReserved) {
            return $this->emptyResponse($this->getResponse(), 500);
        }

        // Get all reservations for current token
        $reservations = $reservationMapper->getAllReservationsByToken($reservationToken);

        $this->addLog($request, $productNumber, $reservationToken, 'Reserved ' . $requestedQuantity . ' successfully. Total reserved is ' . ($requestedQuantity + $currentReservedQuantity));

        return new JsonModel(array(
            'products' => $reservations,
            'reservation_token' => $reservationToken,
            'expiry_datetime' => $reservationTokenExpiry
        ));
    }

    private function addLog(Request $request, $productNumber, $reservationToken, $logMessage) {
        // Get header
        $headers = $request->getHeaders();
        $fromHeader = $headers->get('from');

        // Get header value
        $fromHeaderValue = $fromHeader->getFieldValue();

        // Split header value
        $splitFromHeaderValue = explode('@', $fromHeaderValue);
        $user = $splitFromHeaderValue[0];
        $device = $splitFromHeaderValue[1];

        // Get IP address
        $ipAddress = $request->getServer('REMOTE_ADDR');

        // Create log entity
        $logEntity = new LogEntity();
        $logEntity->setAction($logMessage);
        $logEntity->setDatetime(date('Y-m-d H:i:s'));
        $logEntity->setDevice($device);
        $logEntity->setIpAddress($ipAddress);
        $logEntity->setProductNumber($productNumber);
        $logEntity->setReservationToken($reservationToken);
        $logEntity->setUser($user);

        // Create mapper
        $logMapper = $this->getMapper('LogMapper');

        // Save log
        $logMapper->addLog($logEntity);
    }

    private function emptyResponse(Response $response, $statusCode) {
        $response->setContent('');
        return $response->setStatusCode($statusCode);
    }

    private function getMapper($mapperName) {
        return $this->getServiceLocator()->get($mapperName);
    }

}
